@@ -7,6 +7,7 @@
 
 use crate::alloc::alloc::{alloc, dealloc, Layout};
 use crate::alloc::vec::Vec;
+use crate::bundle::{DynamicBundleClone, DynamicClone};
 use core::any::TypeId;
 use core::ptr::{self, NonNull};
 
@@ -148,7 +149,7 @@ impl Drop for BuiltEntity<'_> {
 /// ```
 #[derive(Clone, Default)]
 pub struct EntityBuilderClone {
-    inner: Common<Cloneable>,
+    inner: Common<DynamicClone>,
 }
 
 impl EntityBuilderClone {
@@ -166,13 +167,25 @@ impl EntityBuilderClone {
             self.inner.add(
                 (&mut component as *mut T).cast(),
                 TypeInfo::of::<T>(),
-                Cloneable::new::<T>(),
+                DynamicClone::new::<T>(),
             );
             core::mem::forget(component);
         }
         self
     }
 
+    /// Add all components in `bundle` to the entity.
+    ///
+    /// If the bundle contains any component which matches the type of a component
+    /// already in the `EntityBuilder`, the newly added component from the bundle
+    /// will replace the old component and the old component will be dropped.
+    pub fn add_bundle(&mut self, bundle: impl DynamicBundleClone) -> &mut Self {
+        unsafe {
+            bundle.put_with_clone(|ptr, ty, cloneable| self.inner.add(ptr, ty, cloneable));
+        }
+        self
+    }
+
     /// Convert into a value whose shared references are [`DynamicBundle`]s suitable for repeated
     /// spawning
     pub fn build(self) -> BuiltEntityClone {
@@ -213,7 +226,7 @@ impl EntityBuilderClone {
 /// Built from, and convertible back into, [`EntityBuilderClone`]. `DynamicBundle` is implemented
 /// for *references to* this type, allowing it to be e.g. spawned repeatedly.
 #[derive(Clone)]
-pub struct BuiltEntityClone(Common<Cloneable>);
+pub struct BuiltEntityClone(Common<DynamicClone>);
 
 unsafe impl DynamicBundle for &'_ BuiltEntityClone {
     fn with_ids<T>(&self, f: impl FnOnce(&[TypeId]) -> T) -> T {
@@ -225,9 +238,18 @@ unsafe impl DynamicBundle for &'_ BuiltEntityClone {
     }
 
     unsafe fn put(self, mut f: impl FnMut(*mut u8, TypeInfo)) {
-        for &(_, offset, ref clone) in &self.0.info {
+        for &(_, offset, clone) in &self.0.info {
             let ptr = self.0.storage.as_ptr().add(offset);
-            (clone.0)(ptr, &mut f);
+            (clone.func)(ptr, &mut f);
+        }
+    }
+}
+
+unsafe impl DynamicBundleClone for &'_ BuiltEntityClone {
+    unsafe fn put_with_clone(self, mut f: impl FnMut(*mut u8, TypeInfo, DynamicClone)) {
+        for &(_, offset, clone) in &self.0.info {
+            let ptr = self.0.storage.as_ptr().add(offset);
+            (clone.func)(ptr, &mut |src, ty| f(src, ty, clone));
         }
     }
 }
@@ -373,21 +395,7 @@ impl<M> Default for Common<M> {
     }
 }
 
-/// Type parameter for [`EntityBuilder`]s with cloneable components
-#[derive(Clone)]
-pub struct Cloneable(unsafe fn(*const u8, &mut dyn FnMut(*mut u8, TypeInfo)));
-
-impl Cloneable {
-    fn new<T: Component + Clone>() -> Self {
-        Self(|src, f| unsafe {
-            let mut tmp = (*src.cast::<T>()).clone();
-            f((&mut tmp as *mut T).cast(), TypeInfo::of::<T>());
-            core::mem::forget(tmp);
-        })
-    }
-}
-
-impl Clone for Common<Cloneable> {
+impl Clone for Common<DynamicClone> {
     fn clone(&self) -> Self {
         unsafe {
             let result = Common {
@@ -399,7 +407,7 @@ impl Clone for Common<Cloneable> {
                 indices: self.indices.clone(),
             };
             for &(_, offset, ref clone) in &self.info {
-                (clone.0)(self.storage.as_ptr().add(offset), &mut |src, ty| {
+                (clone.func)(self.storage.as_ptr().add(offset), &mut |src, ty| {
                     result
                         .storage
                         .as_ptr()
