@@ -9,10 +9,11 @@ use crate::alloc::{vec, vec::Vec};
 use core::any::TypeId;
 use core::borrow::Borrow;
 use core::convert::TryFrom;
+use core::hash::{BuildHasherDefault, Hasher};
 use core::marker::PhantomData;
 use spin::Mutex;
 
-use core::{fmt, mem, ptr};
+use core::{fmt, ptr};
 
 #[cfg(feature = "std")]
 use std::error::Error;
@@ -50,6 +51,12 @@ pub struct World {
     archetypes: ArchetypeSet,
     /// Maps statically-typed bundle types to archetypes
     bundle_to_archetype: TypeIdMap<u32>,
+    /// Maps source archetype and static bundle types to the archetype that an entity is moved to
+    /// after inserting the components from that bundle.
+    insert_edges: IndexTypeIdMap<InsertTarget>,
+    /// Maps source archetype and static bundle types to the archetype that an entity is moved to
+    /// after removing the components from that bundle.
+    remove_edges: IndexTypeIdMap<u32>,
     id: u64,
 }
 
@@ -69,6 +76,8 @@ impl World {
             entities: Entities::default(),
             archetypes: ArchetypeSet::new(),
             bundle_to_archetype: HashMap::default(),
+            insert_edges: HashMap::default(),
+            remove_edges: HashMap::default(),
             id,
         }
     }
@@ -544,25 +553,34 @@ impl World {
         components: impl DynamicBundle,
     ) -> Result<(), NoSuchEntity> {
         self.flush();
-        let loc = self.entities.get_mut(entity)?;
 
+        let loc = self.entities.get(entity)?;
+        self.insert_inner(entity, components, loc.archetype, loc)
+    }
+
+    /// The implementation backing [`insert`](Self::insert) exposed so that it can also be used by [`exchange`](Self::exchange).
+    ///
+    /// Note that `graph_origin` is always equal to `loc.archetype` during insertion. Only for exchange, `graph_origin` identifies
+    /// the intermediate archetype which would be reached after removal and before insertion even though
+    /// the actual component data still resides in `loc.archetype`.
+    fn insert_inner(
+        &mut self,
+        entity: Entity,
+        components: impl DynamicBundle,
+        graph_origin: u32,
+        loc: Location,
+    ) -> Result<(), NoSuchEntity> {
         let target_storage;
         let target = match components.key() {
             None => {
-                target_storage = self
-                    .archetypes
-                    .get_insert_target(loc.archetype, &components);
+                target_storage = self.archetypes.get_insert_target(graph_origin, &components);
                 &target_storage
             }
-            Some(key) => match self.archetypes.insert_edges[loc.archetype as usize].get(&key) {
+            Some(key) => match self.insert_edges.get(&(graph_origin, key)) {
                 Some(x) => x,
                 None => {
-                    let t = self
-                        .archetypes
-                        .get_insert_target(loc.archetype, &components);
-                    self.archetypes.insert_edges[loc.archetype as usize]
-                        .entry(key)
-                        .or_insert(t)
+                    let t = self.archetypes.get_insert_target(graph_origin, &components);
+                    self.insert_edges.entry((graph_origin, key)).or_insert(t)
                 }
             },
         };
@@ -594,8 +612,9 @@ impl World {
 
             // Allocate storage in the archetype and update the entity's location to address it
             let target_index = target_arch.allocate(entity.id);
-            loc.archetype = target.index;
-            let old_index = mem::replace(&mut loc.index, target_index);
+            let meta = &mut self.entities.meta[entity.id as usize];
+            meta.location.archetype = target.index;
+            meta.location.index = target_index;
 
             // Move the new components
             components.put(|ptr, ty| {
@@ -605,14 +624,14 @@ impl World {
             // Move the components we're keeping
             for &ty in &target.retained {
                 let src = source_arch
-                    .get_dynamic(ty.id(), ty.layout().size(), old_index)
+                    .get_dynamic(ty.id(), ty.layout().size(), loc.index)
                     .unwrap();
                 target_arch.put_dynamic(src.as_ptr(), ty.id(), ty.layout().size(), target_index)
             }
 
             // Free storage in the old archetype
-            if let Some(moved) = source_arch.remove(old_index, false) {
-                self.entities.meta[moved as usize].location.index = old_index;
+            if let Some(moved) = source_arch.remove(loc.index, false) {
+                self.entities.meta[moved as usize].location.index = loc.index;
             }
         }
         Ok(())
@@ -662,24 +681,8 @@ impl World {
         };
 
         // Find the target archetype ID
-        let target = match source_arch.remove_edges.get(&TypeId::of::<T>()) {
-            Some(&x) => x,
-            None => {
-                let removed = T::with_static_ids(|ids| ids.iter().copied().collect::<HashSet<_>>());
-                let info = source_arch
-                    .types()
-                    .iter()
-                    .cloned()
-                    .filter(|x| !removed.contains(&x.id()))
-                    .collect::<Vec<_>>();
-                let elements = info.iter().map(|x| x.id()).collect::<Box<_>>();
-                let index = self.archetypes.get(&*elements, move || info);
-                self.archetypes.archetypes[loc.archetype as usize]
-                    .remove_edges
-                    .insert(TypeId::of::<T>(), index);
-                index
-            }
-        };
+        let target =
+            Self::remove_target::<T>(&mut self.archetypes, &mut self.remove_edges, loc.archetype);
 
         // Store components to the target archetype and update metadata
         if loc.archetype != target {
@@ -707,6 +710,31 @@ impl World {
         Ok(bundle)
     }
 
+    fn remove_target<T: Bundle + 'static>(
+        archetypes: &mut ArchetypeSet,
+        remove_edges: &mut IndexTypeIdMap<u32>,
+        old_archetype: u32,
+    ) -> u32 {
+        match remove_edges.get(&(old_archetype, TypeId::of::<T>())) {
+            Some(&x) => x,
+            None => {
+                let removed = T::with_static_ids(|ids| ids.iter().copied().collect::<HashSet<_>>());
+                let info = archetypes.archetypes[old_archetype as usize]
+                    .types()
+                    .iter()
+                    .cloned()
+                    .filter(|x| !removed.contains(&x.id()))
+                    .collect::<Vec<_>>();
+                let elements = info.iter().map(|x| x.id()).collect::<Box<_>>();
+                let index = archetypes.get(&*elements, move || info);
+
+                remove_edges.insert((old_archetype, TypeId::of::<T>()), index);
+
+                index
+            }
+        }
+    }
+
     /// Remove the `T` component from `entity`
     ///
     /// See [`remove`](Self::remove).
@@ -714,6 +742,48 @@ impl World {
         self.remove::<(T,)>(entity).map(|(x,)| x)
     }
 
+    /// Remove `S` components from `entity` and then add `components`
+    ///
+    /// This has the same effect as calling [`remove::<S>`](Self::remove) and then [`insert::<T>`](Self::insert),
+    /// but is more efficient as the intermediate archetype after removal but before insertion is skipped.
+    pub fn exchange<S: Bundle + 'static, T: DynamicBundle>(
+        &mut self,
+        entity: Entity,
+        components: T,
+    ) -> Result<S, ComponentError> {
+        self.flush();
+
+        // Gather current metadata
+        let loc = self.entities.get(entity)?;
+
+        // Move out of the source archetype, or bail out if a component is missing
+        let source_arch = &self.archetypes.archetypes[loc.archetype as usize];
+
+        let bundle = unsafe {
+            S::get(|ty| source_arch.get_dynamic(ty.id(), ty.layout().size(), loc.index))?
+        };
+
+        // Find the intermediate archetype ID
+        let intermediate =
+            Self::remove_target::<S>(&mut self.archetypes, &mut self.remove_edges, loc.archetype);
+
+        self.insert_inner(entity, components, intermediate, loc)?;
+
+        Ok(bundle)
+    }
+
+    /// Remove the `S` component from `entity` and then add `component`
+    ///
+    /// See [`exchange`](Self::exchange).
+    pub fn exchange_one<S: Component, T: Component>(
+        &mut self,
+        entity: Entity,
+        component: T,
+    ) -> Result<S, ComponentError> {
+        self.exchange::<(S,), (T,)>(entity, (component,))
+            .map(|(x,)| x)
+    }
+
     /// Borrow the `T` component of `entity` without safety checks
     ///
     /// Should only be used as a building block for safe abstractions.
@@ -791,6 +861,7 @@ impl World {
     /// let column = world.column::<i32>();
     /// assert_eq!(*column.get(ent).unwrap(), 123);
     /// ```
+    #[deprecated(since = "0.7.2", note = "use the more general QueryBorrow::view")]
     pub fn column<T: Component>(&self) -> Column<'_, T> {
         let archetypes = self.archetypes.archetypes.as_slice();
         let entities = self.entities.meta.as_slice();
@@ -800,6 +871,7 @@ impl World {
     /// Uniquely borrows every `T` component for efficient random access
     ///
     /// See [`World::column`].
+    #[deprecated(since = "0.7.2", note = "use the more general QueryBorrow::view")]
     pub fn column_mut<T: Component>(&self) -> ColumnMut<'_, T> {
         let archetypes = self.archetypes.archetypes.as_slice();
         let entities = self.entities.meta.as_slice();
@@ -1123,10 +1195,6 @@ struct ArchetypeSet {
     index: HashMap<Box<[TypeId]>, u32>,
     archetypes: Vec<Archetype>,
     generation: u64,
-    /// Maps static bundle types to the archetype that an entity from this archetype is moved to
-    /// after inserting the components from that bundle. Stored separately from archetypes to avoid
-    /// borrowck difficulties in `World::insert`.
-    insert_edges: Vec<TypeIdMap<InsertTarget>>,
 }
 
 impl ArchetypeSet {
@@ -1136,7 +1204,6 @@ impl ArchetypeSet {
             index: Some((Box::default(), 0)).into_iter().collect(),
             archetypes: vec![Archetype::new(Vec::new())],
             generation: 0,
-            insert_edges: vec![HashMap::default()],
         }
     }
 
@@ -1193,7 +1260,6 @@ impl ArchetypeSet {
     }
 
     fn post_insert(&mut self) {
-        self.insert_edges.push(HashMap::default());
         self.generation += 1;
     }
 
@@ -1245,6 +1311,29 @@ struct InsertTarget {
     index: u32,
 }
 
+type IndexTypeIdMap<V> = HashMap<(u32, TypeId), V, BuildHasherDefault<IndexTypeIdHasher>>;
+
+#[derive(Default)]
+struct IndexTypeIdHasher(u64);
+
+impl Hasher for IndexTypeIdHasher {
+    fn write_u32(&mut self, index: u32) {
+        self.0 ^= u64::from(index);
+    }
+
+    fn write_u64(&mut self, type_id: u64) {
+        self.0 ^= type_id;
+    }
+
+    fn write(&mut self, _bytes: &[u8]) {
+        unreachable!()
+    }
+
+    fn finish(&self) -> u64 {
+        self.0
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
